/*
 * Copyright (c) 2016-2019,2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Thu Apr 14 2016
 */
#include "acjoystick.h"

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>

#include "joystick_c_types.h"
#include "codels.h"


/* --- Activity log ----------------------------------------------------- */

/** Validation codel joystick_log_open of activity log.
 *
 * Returns genom_ok.
 * Throws joystick_e_sys.
 */
genom_event
joystick_log_open(const char path[64], uint32_t decimation,
                  joystick_log_s **log, const genom_context self)
{
  int fd;

  fd = open(path, O_WRONLY|O_APPEND|O_CREAT|O_TRUNC, 0666);
  if (fd < 0) return joy_e_sys_error(path, self);

  if ((*log)->fd >= 0)
    close((*log)->fd);
  if ((*log)->req.aio_fildes >= 0) {
    if ((*log)->pending)
      while (aio_error(&(*log)->req) == EINPROGRESS)
        /* empty body */;
  }
  (*log)->fd = fd;
  (*log)->req.aio_fildes = -1;
  (*log)->pending = false;
  (*log)->skipped = false;
  (*log)->decimation = decimation < 1 ? 1 : decimation;
  (*log)->missed = 0;
  (*log)->total = 0;

  return genom_ok;
}


/* --- Function device_list --------------------------------------------- */

/** Codel joystick_device_list of function device_list.
 *
 * Returns genom_ok.
 * Throws joystick_e_nomem.
 */
genom_event
joystick_device_list(const sequence_joystick_device_s *devices,
                     sequence_string32 *device_list,
                     const genom_context self)
{
  int i;

  if (genom_sequence_reserve(device_list, devices->_length))
    return joystick_e_nomem(self);

  device_list->_length = devices->_length;
  for(i = 0; i < devices->_length; i++) {
    snprintf(device_list->_buffer[i], sizeof(device_list->_buffer[i]),
             "%s", devices->_buffer[i].name);
  }

  return genom_ok;
}


/* --- Function log_stop ------------------------------------------------ */

/** Codel joystick_log_stop of function log_stop.
 *
 * Returns genom_ok.
 */
genom_event
joystick_log_stop(joystick_log_s **log, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  if (*log && (*log)->req.aio_fildes >= 0)
    close((*log)->req.aio_fildes);
  (*log)->fd = (*log)->req.aio_fildes = -1;

  return genom_ok;
}


/* --- Function log_info ------------------------------------------------ */

/** Codel joystick_log_info of function log_info.
 *
 * Returns genom_ok.
 */
genom_event
joystick_log_info(const joystick_log_s *log, uint32_t *miss,
                  uint32_t *total, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  *miss = *total = 0;
  if (log) {
    *miss = log->missed;
    *total = log->total;
  }
  return genom_ok;
}
