/*
 * Copyright (c) 2016-2019,2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Thu Apr 14 2016
 */

#include "acjoystick.h"

#include <sys/time.h>

#include <err.h>
#include <math.h>
#include <stdio.h>
#include <time.h>

#include "joystick_c_types.h"
#include "codels.h"


/* --- Task state ------------------------------------------------------- */


/** Codel joystick_state_start of task state.
 *
 * Triggered by joystick_start.
 * Yields to joystick_main.
 */
genom_event
joystick_state_start(or_pose_estimator_state *user,
                     const joystick_desired *desired,
                     const genom_context self)
{
  *user = (or_pose_estimator_state){
    .vel._present = true,
    .avel._present = true,
    .acc._present = true,
    .aacc._present = true
  };

  *desired->data(self) = (or_pose_estimator_state){ 0 };
  desired->write(self);

  return joystick_main;
}


/** Codel joystick_state_main of task state.
 *
 * Triggered by joystick_main.
 * Yields to joystick_publish.
 */
genom_event
joystick_state_main(const joystick_params_s *params,
                    const joystick_input_s *input,
                    or_pose_estimator_state *user,
                    const genom_context self)
{
  const double dt = joystick_state_period_ms/1000.;

  or_t3d_vel *v   = &user->vel._value;
  or_t3d_avel *w  = &user->avel._value;
  or_t3d_acc *a   = &user->acc._value;
  or_t3d_aacc *aw = &user->aacc._value;
  struct timeval tv;
  double n2, r;

  /* input */
  a->ax = input->a.ax - params->drag * v->vx;
  a->ay = input->a.ay - params->drag * v->vy;
  a->az = input->a.az - params->drag * v->vz;
  n2 = a->ax * a->ax + a->ay * a->ay + a->az * a->az;
  if (n2 > params->amax * params->amax) {
    r = params->amax/sqrt(n2);
    a->ax *= r;
    a->ay *= r;
    a->az *= r;
  }

  aw->awx = input->aw.awx - params->drag * w->wx;
  aw->awy = input->aw.awy - params->drag * w->wy;
  aw->awz = input->aw.awz - params->drag * w->wz;
  n2 = aw->awx * aw->awx + aw->awy * aw->awy + aw->awz * aw->awz;
  if (n2 > params->awmax * params->awmax) {
    r = params->awmax/sqrt(n2);
    aw->awx *= r;
    aw->awy *= r;
    aw->awz *= r;
  }


  /* integration */
  v->vx += dt*a->ax;
  v->vy += dt*a->ay;
  v->vz += dt*a->az;

  w->wx += dt*aw->awx;
  w->wy += dt*aw->awy;
  w->wz += dt*aw->awz;

  /* flags */
  gettimeofday(&tv, NULL);
  user->ts.sec = tv.tv_sec; user->ts.nsec = tv.tv_usec * 1000;
  user->intrinsic = input->intrinsic;

  return joystick_publish;
}


/** Codel joystick_state_publish of task state.
 *
 * Triggered by joystick_publish.
 * Yields to joystick_pause_main, joystick_log.
 */
genom_event
joystick_state_publish(const joystick_input_s *input,
                       const or_pose_estimator_state *user,
                       const joystick_reference *reference,
                       const joystick_desired *desired,
                       const genom_context self)
{
  or_pose_estimator_state *s = desired->data(self);
  or_pose_estimator_state *r;
  struct timeval tv;

  if (input->active)
    *s = *user;
  else if (reference->read(self) == genom_ok &&
           (r = reference->data(self)))
    *s = *r;
  else if (s->vel._present) {
    gettimeofday(&tv, NULL);
    *s = (or_pose_estimator_state){
      .ts = { .sec = tv.tv_sec, .nsec = tv.tv_usec * 1000 }
    };
  } else return joystick_pause_main;

  desired->write(self);
  return joystick_log;
}


/** Codel joystick_state_log of task state.
 *
 * Triggered by joystick_log.
 * Yields to joystick_pause_main.
 */
genom_event
joystick_state_log(const joystick_input_s *input, joystick_log_s **log,
                   const joystick_desired *desired,
                   const genom_context self)
{
  or_pose_estimator_state *s = desired->data(self);
  int n;

  if ((*log)->req.aio_fildes < 0) return joystick_pause_main;

  (*log)->total++;
  if ((*log)->total % (*log)->decimation) return joystick_pause_main;

  if ((*log)->pending) {
    switch (aio_error(&(*log)->req)) {
      case EINPROGRESS:
        (*log)->skipped = true;
        (*log)->missed++;
        return joystick_pause_main;
    }

    (*log)->pending = false;
    if (aio_return(&(*log)->req) <= 0) {
      warn("log");
      joystick_log_stop(log, self);
      return joystick_pause_main;
    }
  }

  /* build log buffer */
  n = snprintf(
    (*log)->buffer, sizeof((*log)->buffer),
    "%s"
    "%"PRIu64".%09d %d "
    " %g %g %g  %g %g %g  %g %g %g  %g %g %g\n",
    (*log)->skipped ? "\n" : "",
    (uint64_t)s->ts.sec, (uint32_t)s->ts.nsec, input->active,
    s->vel._value.vx, s->vel._value.vy, s->vel._value.vz,
    s->avel._value.wx, s->avel._value.wy, s->avel._value.wz,
    s->acc._value.ax, s->acc._value.ay, s->acc._value.az,
    s->aacc._value.awx, s->aacc._value.awy, s->aacc._value.awz);
  if (n < 0 || n >= sizeof((*log)->buffer)) goto full;

  (*log)->req.aio_nbytes = n;
  if (aio_write(&(*log)->req)) {
    warn("log");
    goto err;
  }

  (*log)->pending = true;
  (*log)->skipped = false;
  return joystick_pause_main;

full:
  warnx("log buffer overflow");
err:
  joystick_log_stop(log, self);
  return joystick_pause_main;
}


/** Codel joystick_state_stop of task state.
 *
 * Triggered by joystick_stop.
 * Yields to joystick_ether.
 */
genom_event
joystick_state_stop(const joystick_desired *desired,
                    const genom_context self)
{
  or_pose_estimator_state *s = desired->data(self);

  *s = (or_pose_estimator_state){ 0 };
  desired->write(self);

  return joystick_ether;
}


/* --- Activity log ----------------------------------------------------- */

/** Codel joystick_log_header of activity log.
 *
 * Triggered by joystick_start.
 * Yields to joystick_ether.
 * Throws joystick_e_sys.
 */
genom_event
joystick_log_header(joystick_log_s **log, const genom_context self)
{
  int s;

  s = dprintf(
    (*log)->fd, "# logged on %s#\n" /* note that ctime(3) has a \n */,
    ctime(&(time_t){ time(NULL) }));
  if (s < 0) goto err;

  s = dprintf((*log)->fd, joy_log_header_fmt "\n");
  if (s < 0) goto err;

  /* enable async writes */
  (*log)->req.aio_fildes = (*log)->fd;

  return joystick_ether;
err:
  joystick_log_stop(log, self);
  return joy_e_sys_error("log", self);
}


/* --- joystick_mkstateflags ----------------------------------------------- */

void
joystick_mkstateflags(int button, int pressed,
                      const joystick_mappings_s *mappings,
                      const or_joystick_state *state, joystick_input_s *input)
{
  if (pressed) {
    if (button == mappings->coords.intrinsic)
      input->intrinsic = true;
    else if (button == mappings->coords.world)
      input->intrinsic = false;
    else if (button == mappings->coords.toggle)
      input->intrinsic = !input->intrinsic;

    else if (button == mappings->enable.on)
      input->active = true;
    else if (button == mappings->enable.slow)
      input->active = true;
    else if (button == mappings->enable.toggle)
      input->active = !input->active;

    else if (mappings->enable.toggle < 0 &&
             mappings->enable.slow < 0 &&
             mappings->enable.on < 0)
      input->active = true;
  } else {
    if (button == mappings->enable.on &&
        (mappings->enable.slow < 0 ||
         !state->buttons._buffer[mappings->enable.slow]))
      input->active = false;
    else if (button == mappings->enable.slow &&
             (mappings->enable.on < 0 ||
              !state->buttons._buffer[mappings->enable.on]))
      input->active = false;
  }
}


/* --- joystick_mkmotion --------------------------------------------------- */

void
joystick_mkmotion(const joystick_params_s *params,
                  const joystick_mappings_s *mappings,
                  const or_joystick_state *state, joystick_input_s *input)
{
  double v, scale;
  int i;

  switch(
    (mappings->enable.on < 0 ||
     state->buttons._buffer[mappings->enable.on]) |
    ((mappings->enable.slow >= 0 &&
      state->buttons._buffer[mappings->enable.slow]) << 1)) {
    case 0: scale = 0.; break;
    case 1: scale = 1.; break;
    case 3: scale = 0.5 + 1. / params->slow / 2.; break;
    case 2: scale = 1. / params->slow; break;
  }

  for(i = 0; i < 3; i++) {
    if (mappings->linear[i].axis < 0) v = 0.;
    else
      v = scale * mappings->linear[i].scale * params->amax *
          (state->axes._buffer[mappings->linear[i].axis] / 32768. -
           mappings->linear[i].offset);
    switch(i) {
      case 0: input->a.ax = v; break;
      case 1: input->a.ay = v; break;
      case 2: input->a.az = v; break;
    }
  }

  for(i = 0; i < 3; i++) {
    if (mappings->angular[i].axis < 0) v = 0.;
    else
      v = scale * mappings->angular[i].scale * params->awmax *
          (state->axes._buffer[mappings->angular[i].axis] / 32768. -
           mappings->angular[i].offset);
    switch(i) {
      case 0: input->aw.awx = v; break;
      case 1: input->aw.awy = v; break;
      case 2: input->aw.awz = v; break;
    }
  }
}
